from typing import Optional
from typeguard import typechecked
import os


@typechecked
class Env():
    '''
    Task Environment definition
    '''

    def __init__(
        self, name: str,
        os_name: Optional[str] = None,
        default: str = ''
    ):
        self.name: str = name
        self.os_name: str = os_name if os_name is not None else name
        self.default: str = default

    def get(self, prefix: str = '') -> str:
        '''
        Return environment value.
        You can use prefix to distinguish 'DEV', 'PROD'

        Example:
        ```python
        os.environ['DEV_SERVER'] = 'localhost'
        os.environ['PROD_SERVER'] = 'example.com'

        env = Env(name='HOST', os_name='SERVER', default='0.0.0.0')

        print(env.get('DEV'))   # will show 'localhost'
        print(env.get('PROD'))  # will show 'example.com'
        print(env.get('STAG'))  # will show '0.0.0.0'
        ```
        '''
        prefixed_name = self._get_prefixed_name(self.os_name, prefix)
        return os.getenv(prefixed_name, os.getenv(self.os_name, self.default))

    def _get_prefixed_name(self, name: str, prefix: str):
        if prefix is None or prefix == '':
            return name
        return '_'.join([prefix, name])



from typing import Any, List


def append_unique(data: List[Any], *new_data: Any) -> List[Any]:
    for new_datum in new_data:
        if new_datum in data:
            continue
        data.append(new_datum)
    return data



from .helper.cli.create_cli import create_cli
from .config.config import logging_level
from termcolor import colored
import logging

logging.basicConfig(
    level=logging_level,
    format=colored('%(levelname)s', attrs=['dark']) + ' %(message)s'
)
cli = create_cli()


if __name__ == '__main__':
    cli()



from typing import Any, Callable, List, Mapping, Optional
from typeguard import typechecked
from .base_task import BaseTask
from ..task_env.env import Env
from ..task_group.group import Group
from ..task_input.base_input import BaseInput
from ..helper.file.copy_tree import copy_tree


@typechecked
class CodeMaker(BaseTask):

    def __init__(
        self,
        name: str,
        template_path: str,
        destination_path: str,
        replacements: Mapping[str, str] = {},
        excludes: List[str] = [],
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: List[BaseTask] = [],
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=0.1,
            retry=0,
            retry_interval=0
        )
        self.template_path = template_path
        self.destination_path = destination_path
        self.excudeds = excludes
        rendered_replacements: Mapping[str, str] = {
            old: self.render_str(new)
            for old, new in replacements.items()
        }
        self.replacements = rendered_replacements

    def create_main_loop(
        self, env_prefix: str = ''
    ) -> Callable[..., bool]:
        return super().create_main_loop(env_prefix)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        copy_tree(
            src=self.template_path,
            dst=self.destination_path,
            replacements=self.replacements,
            excludes=self.excudeds
        )
        return True



from ...runner import runner
from ..loader.load_module import load_module
from ...config.config import init_scripts, should_load_default
from ...default._register import register_default

import click
import os


def create_cli() -> click.Group:
    cli = click.Group(name='zrb', help='Your faithful companion.')

    if 'PYTHONUNBUFFERED' not in os.environ:
        os.environ['PYTHONUNBUFFERED'] = '1'

    # load from ZRB_INIT_SCRIPTS environment
    for init_script in init_scripts:
        load_module(script_path=init_script)

    # Load zrb_init.py
    project_script = os.path.join(os.getcwd(), 'zrb_init.py')
    if os.path.isfile(project_script):
        load_module(script_path=project_script)

    if should_load_default:
        register_default(runner)

    # Serve all tasks registered to runner
    cli = runner.serve(cli)
    return cli



from zrb.task.cmd_task import CmdTask
import pathlib
import os


def test_simple_with_no_error():
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello'
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == 'hello'


def test_simple_with_error():
    cmd_task = CmdTask(
        name='simple-error',
        cmd='forbidden command'
    )
    main_loop = cmd_task.create_main_loop()
    is_error: bool = False
    try:
        main_loop()
    except Exception:
        is_error = True
    assert is_error


def test_multiline_cmd():
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd=[
            'echo hello',
            'echo hello again',
        ]
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == '\n'.join(['hello', 'hello again'])


def test_cmd_path():
    dir_path = pathlib.Path(__file__).parent.absolute()
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd_path=os.path.join(dir_path, 'templates', 'hello.sh')
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == 'Hello, World!'


def test_upstream_with_no_error():
    upstream_task = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_task]
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == 'hello'


def test_upstream_with_error():
    upstream_task = CmdTask(
        name='upstream-error',
        cmd='forbidden command'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_task]
    )
    main_loop = cmd_task.create_main_loop()
    is_error: bool = False
    try:
        main_loop()
    except Exception:
        is_error = True
    assert is_error


def test_diamond_upstream():
    upstream_0 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 0'
    )
    upstream_1 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 1',
        upstreams=[upstream_0]
    )
    upstream_2 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 2',
        upstreams=[upstream_0]
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_1, upstream_2]
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == 'hello'


def test_checker_with_no_error():
    checker = CmdTask(
        name='check',
        cmd='echo checked'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        checkers=[checker]
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == 'hello'


def test_checker_with_error():
    checker = CmdTask(
        name='check',
        cmd='forbidden command'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        checkers=[checker]
    )
    main_loop = cmd_task.create_main_loop()
    is_error: bool = False
    try:
        main_loop()
    except Exception:
        is_error = True
    assert is_error


def test_long_output():
    cmd_task = CmdTask(
        name='long-output',
        cmd=[
            'echo 1',
            'echo 2',
            'echo 3',
            'echo 4',
            'echo 5',
        ],
        max_output_line=3,
    )
    main_loop = cmd_task.create_main_loop()
    result = main_loop()
    assert result.output == '\n'.join(['3', '4', '5'])



import logging
import os


def get_logging_level():
    logging_level_str = os.getenv('ZRB_LOGGING_LEVEL', 'WARNING').lower()
    logging_level_map = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'warn': logging.WARN,
        'info': logging.INFO,
        'debug': logging.DEBUG,
        'notset': logging.NOTSET,
    }
    if logging_level_str in logging_level_map:
        return logging_level_map[logging_level_str]
    return logging.WARNING


init_scripts = os.getenv('ZRB_INIT_SCRIPTS', '').split(':')
logging_level = get_logging_level()
should_load_default = os.getenv('ZRB_SHOULD_LOAD_DEFAULT', '1') != '0'
env_prefix = os.getenv('ZRB_ENV', '')



from typing import Any, Callable, List, Optional, Union
from typeguard import typechecked
from http.client import HTTPConnection, HTTPSConnection
from .base_task import BaseTask
from ..task_env.env import Env
from ..task_group.group import Group
from ..task_input.base_input import BaseInput

import asyncio


@typechecked
class HTTPChecker(BaseTask):

    def __init__(
        self,
        name: str = 'http_check',
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        host: str = 'localhost',
        port: Union[int, str] = 80,
        timeout: Union[int, str] = 5,
        method: str = 'HEAD',
        url: str = '/',
        is_https: bool = False,
        upstreams: List[BaseTask] = [],
        checking_interval: float = 0.1,
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=checking_interval,
            retry=0,
            retry_interval=0
        )
        self.host = host
        self.port = port
        self.timeout = timeout
        self.method = method
        self.url = url
        self.is_https = is_https

    def create_main_loop(
        self, env_prefix: str = ''
    ) -> Callable[..., bool]:
        return super().create_main_loop(env_prefix)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        method = self.render_str(self.method)
        host = self.render_str(self.host)
        port = self.get_int(self.port)
        url = self.render_str(self.url)
        timeout = self.get_int(self.timeout)
        while not self._check_connection(method, host, port, url, timeout):
            await asyncio.sleep(self.checking_interval)
        return True

    def _check_connection(
        self, method: str, host: str, port: int, url: str, timeout: int
    ) -> bool:
        label = self._get_label(method, host, port, url)
        conn = self._get_connection(host, port, timeout)
        try:
            conn.request(method, url)
            res = conn.getresponse()
            if res.status < 300:
                self.log_info('Connection success')
                self.print_out(f'{label} {res.status} (OK)')
                return True
            self.log_debug(f'{label} {res.status} (Not OK)')
        except Exception:
            self.log_debug(f'{label} Connection error')
        finally:
            conn.close()
        return False

    def _get_label(self, method: str, host: str, port: int, url: str) -> str:
        protocol = 'https' if self.is_https else 'http'
        return f'{method} {protocol}://{host}:{port}{url}'

    def _get_connection(
        self, host: str, port: int, timeout: int
    ) -> Union[HTTPConnection, HTTPSConnection]:
        if self.is_https:
            return HTTPSConnection(host, port, timeout=timeout)
        return HTTPConnection(host, port, timeout=timeout)



from typing import Any, Callable, List, Optional, Union
from typeguard import typechecked
from .base_task import BaseTask
from ..task_env.env import Env
from ..task_group.group import Group
from ..task_input.base_input import BaseInput

import socket
import asyncio


@typechecked
class PortChecker(BaseTask):

    def __init__(
        self,
        name: str = 'port_check',
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        host: str = 'localhost',
        port: Union[int, str] = 80,
        timeout: Union[int, str] = 5,
        upstreams: List[BaseTask] = [],
        checking_interval: float = 0.1,
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=checking_interval,
            retry=0,
            retry_interval=0
        )
        self.host = host
        self.port = port
        self.timeout = timeout

    def create_main_loop(
        self, env_prefix: str = ''
    ) -> Callable[..., bool]:
        return super().create_main_loop(env_prefix)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        host = self.render_str(self.host)
        port = self.get_int(self.port)
        timeout = self.get_int(self.timeout)
        while not self._check_port(host, port, timeout):
            await asyncio.sleep(self.checking_interval)
        return True

    def _check_port(
        self, host: str, port: int, url: str, timeout: int
    ) -> bool:
        label = self._get_label(host, port, url)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(timeout)
            result = sock.connect_ex((host, port))
            if result == 0:
                self.print_out(f'{label} (OK)')
                return True
        self.log_debug(f'{label} (Not OK)')
        return False

    def _get_label(self, host: str, port: int) -> str:
        return f'Checking {host}:{port}'



from typing import Any
from zrb import (
    runner, Env, StrInput, Group, Task, CmdTask, HTTPChecker
)

'''
Simple Python task, concatenate words
'''
concat = Task(
    name='concat',
    inputs=[StrInput(name='separator', description='Separator', default=' ')],
)
runner.register(concat)


@concat.runner
def run(*args: str, **kwargs: Any) -> str:
    separator = kwargs.get('separator', ' ')
    return separator.join(args)


'''
Simple CLI task, read input and show output
'''
hello = CmdTask(
    name='hello',
    inputs=[StrInput(name='name', description='Name', default='world')],
    cmd='echo Hello {{input.name}}'
)
runner.register(hello)

make = Group(name='make', description='Make things')

'''
Simple CLI task, part of 'make' group
'''
make_coffee = CmdTask(
    name='coffee',
    group=make,
    upstreams=[hello],
    cmd='echo Coffee for you ☕'
)
runner.register(make_coffee)

'''
Simple CLI task, part of 'make' group
'''
make_beer = CmdTask(
    name='beer',
    group=make,
    upstreams=[hello],
    cmd='echo Cheers 🍺'
)
runner.register(make_beer)

'''
Sub group of 'make'
'''
make_gitignore = Group(
    name='gitignore', description='Make gitignore', parent=make
)

'''
Simple CLI task, part of 'make_gitignore' group.
Having multiline cmd
'''
make_gitignore_python = CmdTask(
    name='node',
    group=make_gitignore,
    cmd=[
        'echo "node_modules/" >> .gitignore'
        'echo ".npm" >> .gitignore'
        'echo "npm-debug.log" >> .gitignore'
    ]
)
runner.register(make_gitignore_python)

'''
Simple CLI task, part of 'make_gitignore' group.
Having multiline cmd
'''
make_gitignore_nodejs = CmdTask(
    name='node',
    group=make_gitignore,
    cmd=[
        'echo "__pycache__/" >> .gitignore'
        'echo "venv" >> .gitignore'
    ]
)
runner.register(make_gitignore_nodejs)

server = Group(
    name='server', description='Server related commands'
)

'''
Long running CLI task.
Run a server and waiting for the port to be ready.
'''
run_server = CmdTask(
    name='run',
    group=server,
    upstreams=[make_coffee, make_beer],
    inputs=[StrInput(name='dir', description='Directory', default='.')],
    envs=[Env(name='PORT', os_name='WEB_PORT', default='3000')],
    cmd='python -m http.server $PORT --directory {{input.dir}}',
    checkers=[HTTPChecker(port='{{env.PORT}}')]
)
runner.register(run_server)


from typing import List, Mapping
from ..string.parse_replacement import parse_replacement

import os
import shutil
import fnmatch


def copy_tree(
    src: str,
    dst: str,
    replacements: Mapping[str, str],
    excludes: List[str] = []
):
    names = os.listdir(src)
    new_dst = parse_replacement(dst, replacements)
    if not os.path.exists(new_dst):
        os.makedirs(new_dst)
    for name in names:
        src_name = os.path.join(src, name)
        if any(fnmatch.fnmatch(src_name, pattern) for pattern in excludes):
            continue
        dst_name = os.path.join(dst, name)
        if os.path.isdir(src_name):
            copy_tree(src_name, dst_name, replacements, excludes)
            continue
        new_dst_name = parse_replacement(dst_name, replacements)
        shutil.copy2(src_name, new_dst_name)
        with open(new_dst_name, 'r') as file:
            file_content = file.read()
        new_file_content = parse_replacement(file_content, replacements)
        with open(new_dst_name, 'w') as file:
            file.write(new_file_content)



from typing import Any, Callable, List, Mapping, Optional, Type, Union
from typeguard import typechecked


@typechecked
class BaseInput():

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Any = None,
        description: Optional[str] = None,
        show_default: Union[bool, str, None] = None,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        type: Optional[Any] = None,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1,
    ):
        self.name = name
        self.shortcut = shortcut
        self.prompt = prompt
        self.default = default
        self.help = description if description is not None else name
        self.type = type
        self.show_default = show_default
        self.confirmation_prompt = confirmation_prompt
        self.prompt_required = prompt_required
        self.hide_input = hide_input
        self.is_flag = is_flag
        self.flag_value = flag_value
        self.multiple = multiple
        self.count = count
        self.allow_from_autoenv = allow_from_autoenv
        self.hidden = hidden
        self.show_choices = show_choices
        self.show_envvar = show_envvar
        self.nargs = nargs

    def get_name(self) -> str:
        return self.name

    def get_param_decl(self) -> List[str]:
        param_decl: List[str] = [f'--{self.name}']
        if self.shortcut is not None:
            param_decl.append(f'-{self.shortcut}')
        return param_decl

    def get_options(self) -> Mapping[str, Any]:
        return {
            'prompt': self.prompt,
            'default': self.default,
            'help': self.help,
            'type': self.type,
            'show_default': self.show_default,
            'confirmation_prompt': self.confirmation_prompt,
            'prompt_required': self.prompt_required,
            'hide_input': self.hide_input,
            'is_flag': self.is_flag,
            'flag_value': self.flag_value,
            'multiple': self.multiple,
            'count': self.count,
            'allow_from_autoenv': self.allow_from_autoenv,
            'hidden': self.hidden,
            'show_choices': self.show_choices,
            'show_envvar': self.show_envvar,
            'nargs': self.nargs
        }



from typing import Any, Mapping


class DynamicModel:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return str(self.__dict__)


def get_object_from_keyval(map: Mapping[str, Any]) -> DynamicModel:
    return DynamicModel(**map)



import random
import string


def get_random_name(
    separator: str = '-',
    add_random_digit: bool = True,
    digit_count: int = 3
) -> str:
    prefixes = [
        "aurum", "argentum", "platinum", "mercurius", "sulfur", "sal",
        "luna", "sol", "ferrum", "cuprum"
    ]
    suffixes = [
        "philosophorum", "spiritus", "tinctura", "essentia", "elixir",
        "praeparatum", "aether", "vitae", "lapis", "metallum"
    ]
    prefix = random.choice(prefixes)
    suffix = random.choice(suffixes)
    parts = [prefix, suffix]
    if add_random_digit:
        random_digit = ''.join(
            random.choices(string.digits, k=digit_count)
        )
        parts.append(random_digit)
    return separator.join(parts)



from typing import Optional, TypeVar
from typeguard import typechecked
from ..helper.string.get_cmd_name import get_cmd_name

TGroup = TypeVar('TGroup', bound='Group')


@typechecked
class Group():

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        parent: Optional[TGroup] = None
    ):
        self.name = name
        self.description = description
        self.parent = parent

    def get_cmd_name(self) -> str:
        return get_cmd_name(self.name)

    def get_complete_name(self) -> str:
        cmd_name = self.get_cmd_name()
        if self.parent is None:
            return cmd_name
        parent_cmd_name = self.parent.get_complete_name()
        return f'{parent_cmd_name} {cmd_name}'

    def get_id(self) -> str:
        group_id = self.get_cmd_name()
        if self.parent is None:
            return group_id
        parent_group_id = self.parent.get_id()
        return f'{parent_group_id} {group_id}'



from typing import Any, Callable, List, Mapping, Optional, TypeVar
from typeguard import typechecked
from .base_model import TaskModel
from ..task_input.base_input import BaseInput
from ..task_env.env import Env
from ..task_group.group import Group
from ..helper.list.append_unique import append_unique

import asyncio
import copy
import inspect

TTask = TypeVar('TTask', bound='BaseTask')


@typechecked
class BaseTask(TaskModel):
    '''
    Base class for all tasks.
    Every task definition should be extended from this class.
    '''

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: List[TTask] = [],
        checkers: List[TTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
    ):
        TaskModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            icon=icon,
            color=color,
            retry=retry
        )
        self.inputs = inputs
        self.description = description
        self.retry_interval = self.sanitize_interval(
            retry_interval, 'retry'
        )
        self.upstreams = upstreams
        self.checkers = checkers
        self.checking_interval = self.sanitize_interval(
            checking_interval, 'checking'
        )
        self._is_checked: bool = False
        self._is_executed: bool = False
        self._runner: Optional[Callable[..., Any]] = None

    def runner(self, runner: Callable[..., Any]):
        self._runner = runner

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        '''
        Do task execution
        Please override this method.
        '''
        self.log_debug(
            f'Run with args: {args} and kwargs: {kwargs}'
        )
        if self._runner is not None:
            result = self._runner(*args, **kwargs)
            if inspect.isawaitable(result):
                sync_result = await result
                self.print_out(sync_result)
                return sync_result
            self.print_out(result)
            return result
        return True

    async def check(self) -> bool:
        '''
        Return true when task is considered completed.
        By default, this will wait the task execution to be completed.
        You can override this method.
        '''
        return self.is_done()

    def get_all_inputs(self) -> List[BaseInput]:
        ''''
        Getting all inputs of this task and all its upstream, non-duplicated.
        '''
        inputs: List[BaseInput] = []
        for upstream in self.upstreams:
            upstream_inputs = upstream.get_all_inputs()
            append_unique(inputs, *upstream_inputs)
        append_unique(inputs, *self.inputs)
        return inputs

    def get_description(self) -> str:
        if self.description != '':
            return self.description
        return self.name

    def sanitize_interval(self, interval: float, label: str) -> float:
        if interval < 0:
            name = self._get_complete_name()
            self.log_warn(
                f'Find negative {label} interval for {name}: {interval}'
            )
            return 0
        return interval

    def create_main_loop(self, env_prefix: str = '') -> Callable[..., Any]:
        self_cp = copy.deepcopy(self)

        def main_loop(**kwargs: Any) -> Any:
            '''
            Task main loop.
            '''
            async def run_and_check_all_async() -> Any:
                self_cp._set_keyval(input_map=kwargs, env_prefix=env_prefix)
                processes = [
                    asyncio.create_task(self_cp._loop_check(celebrate=True)),
                    asyncio.create_task(self_cp._run_all(**kwargs))
                ]
                results = await asyncio.gather(*processes)
                return results[-1]
            try:
                return asyncio.run(run_and_check_all_async())
            except Exception:
                self_cp.log_error('Failed')
                raise
            finally:
                self_cp.play_bell()
        return main_loop

    async def _loop_check(self, celebrate: bool = False) -> bool:
        while not await self._cached_check():
            self.log_debug('Task is not ready')
            await asyncio.sleep(self.checking_interval)
        self.end_timer()
        self.log_info('Task is ready')
        if celebrate:
            self.show_celebration()
        return True

    async def _cached_check(self) -> bool:
        if self._is_checked:
            self.log_debug('Skip checking, because checking flag has been set')
            return True
        check_result = await self._check()
        if check_result:
            self._is_checked = True
            self.log_debug('Set checking flag')
        return check_result

    async def _check(self) -> bool:
        '''
        Check current task readiness.
        - If self.checkers is defined,
          this will return True once every self.checkers is completed
        - Otherwise, this will return check method's return value.
        '''
        if len(self.checkers) == 0:
            return await self.check()
        check_processes: List[asyncio.Task] = []
        for checker_task in self.checkers:
            check_processes.append(asyncio.create_task(checker_task.run()))
        await asyncio.gather(*check_processes)
        return True

    async def _run_all(self, **kwargs: Any) -> Any:
        processes: List[asyncio.Task] = []
        # Add upstream tasks to processes
        for upstream_task in self.upstreams:
            processes.append(asyncio.create_task(
                upstream_task._run_all(**kwargs)
            ))
        # Add current task to processes
        processes.append(self._cached_run(**kwargs))
        # Wait everything to complete
        results = await asyncio.gather(*processes)
        return results[-1]

    async def _cached_run(self, **kwargs: Any) -> Any:
        if self._is_executed:
            self.log_debug('Skip running, because execution flag has been set')
            return
        self._is_executed = True
        self.log_debug('Set execution flag and running')
        self.start_timer()
        # get upstream checker
        upstream_check_processes: List[asyncio.Task] = []
        for upstream_task in self.upstreams:
            upstream_check_processes.append(asyncio.create_task(
                upstream_task._loop_check()
            ))
        # wait all upstream checkers to complete
        await asyncio.gather(*upstream_check_processes)
        # initiate args
        args: List[Any] = [] if '_args' not in kwargs else kwargs['_args']
        # start running task
        result: Any
        while self.should_attempt():
            try:
                result = await self.run(*args, **kwargs)
                break
            except Exception:
                if self.is_last_attempt():
                    raise
                attempt = self.get_attempt()
                self.log_error(f'Encounter error on attempt {attempt}')
                self.increase_attempt()
                await asyncio.sleep(self.retry_interval)
        self.mark_as_done()
        return result

    def _set_keyval(self, input_map: Mapping[str, Any], env_prefix: str):
        self.set_local_keyval(input_map=input_map, env_prefix=env_prefix)
        for upstream_task in self.upstreams:
            upstream_task._set_keyval(
                input_map=input_map, env_prefix=env_prefix
            )
        local_env_map = self.get_env_map()
        for checker_task in self.checkers:
            checker_task._set_keyval(
                input_map=input_map, env_prefix=env_prefix
            )
            # Checker task should be able to read local env
            checker_task.inject_env_map(local_env_map)



from typing import Any, Callable, Iterable, List, Optional, Union
from typeguard import typechecked
from .base_task import BaseTask
from ..task_env.env import Env
from ..task_input.base_input import BaseInput
from ..task_group.group import Group

import asyncio


class CmdResult():
    def __init__(self, output: str, error: str):
        self.output = output
        self.error = error


@typechecked
class CmdTask(BaseTask):
    '''
    Command Task.
    You can use this task to run shell command.

    For example:
    ```python
    # run a simple task
    hello = CmdTask(
        name='hello',
        inputs=[StrInput(name='name', default='World')],
        envs=[Env(name='HOME_DIR', os_name='HOME')],
        cmd=[
            'echo Hello {{ input.name }}',
            'echo Home directory is: $HOME_DIR',
        ]
    )
    runner.register(hello)

    # run a long running process
    run_server = CmdTask(
        name='run',
        inputs=[StrInput(name='dir', default='.')],
        envs=[Env(name='PORT', os_name='WEB_PORT', default='3000')],
        cmd='python -m http.server $PORT --directory {{input.dir}}',
        checkers=[HTTPChecker(port='{{env.PORT}}')]
    )
    runner.register(run_server)
    ```
    '''

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        cmd: Union[str, Iterable[str]] = '',
        cmd_path: str = '',
        upstreams: List[BaseTask] = [],
        checkers: List[BaseTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval
        )
        self.cmd = cmd
        self.cmd_path = cmd_path
        self.max_output_size = max_output_line
        self.max_error_size = max_error_line
        self._output_buffer: List[str] = []
        self._error_buffer: List[str] = []

    def create_main_loop(
        self, env_prefix: str = ''
    ) -> Callable[..., CmdResult]:
        return super().create_main_loop(env_prefix)

    async def run(self, *args: Any, **kwargs: Any) -> CmdResult:
        cmd = self._get_cmd_str()
        env = self.get_env_map()
        self.log_debug(f'Run command: {cmd}\nwith env: {env}')
        self._output_buffer = []
        self._error_buffer = []
        process = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env,
            shell=True
        )
        self.set_task_pid(process.pid)
        await self._wait_process(process)
        # get output and error
        output = '\n'.join(self._output_buffer)
        error = '\n'.join(self._error_buffer)
        # get return code
        return_code = process.returncode
        if return_code != 0:
            self.log_debug(f'Exit status: {return_code}')
            raise Exception(
                f'Process {self.name} exited ({return_code}): {error}'
            )
        return CmdResult(output, error)

    async def _wait_process(self, process: asyncio.subprocess.Process):
        # Create queue
        stdout_queue = asyncio.Queue()
        stderr_queue = asyncio.Queue()
        # Read from streams and put into queue
        stdout_process = asyncio.create_task(self._queue_stream(
            process.stdout, stdout_queue
        ))
        stderr_process = asyncio.create_task(self._queue_stream(
            process.stderr, stderr_queue
        ))
        # Handle messages in queue
        stdout_log_process = asyncio.create_task(self._log_from_queue(
            stdout_queue, self.print_out,
            self._output_buffer, self.max_output_size
        ))
        stderr_log_process = asyncio.create_task(self._log_from_queue(
            stderr_queue, self.print_err,
            self._error_buffer, self.max_error_size
        ))
        # wait process
        await process.wait()
        # wait reader and logger
        await stdout_process
        await stderr_process
        await stdout_queue.put(None)
        await stderr_queue.put(None)
        await stdout_log_process
        await stderr_log_process

    def _get_cmd_str(self) -> str:
        if self.cmd_path != '':
            return self.render_file(self.cmd_path)
        if isinstance(self.cmd, str):
            return self.render_str(self.cmd)
        return self.render_str('\n'.join(self.cmd))

    async def _queue_stream(self, stream, queue: asyncio.Queue):
        while True:
            line = await stream.readline()
            if not line:
                break
            await queue.put(line)

    async def _log_from_queue(
        self,
        queue: asyncio.Queue,
        print_log: Callable[[str], None],
        buffer: List[str],
        max_size: int
    ):
        while True:
            line = await queue.get()
            if not line:
                break
            line_str = line.decode().rstrip()
            self._add_to_buffer(buffer, max_size, line_str)
            print_log(line_str)

    def _add_to_buffer(
        self, buffer: List[str], max_size: int, new_line: str
    ):
        if len(buffer) >= max_size:
            buffer.pop(0)
        buffer.append(new_line)



from typing import Any, List, Mapping, Optional, Union
from typeguard import typechecked
from ..helper.accessories.color import (
    get_random_color, is_valid_color, colored
)
from ..helper.accessories.icon import get_random_icon
from ..helper.keyval.get_object_from_keyval import get_object_from_keyval
from ..helper.string.get_cmd_name import get_cmd_name
from ..task_env.env import Env
from ..task_group.group import Group

import datetime
import logging
import os
import sys
import time
import jinja2


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                    return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class TimeTracker():

    def __init__(self):
        self._start_time: float = 0
        self._end_time: float = 0

    def reset_timer(self):
        self._start_time = 0
        self._end_time = 0

    def start_timer(self):
        self._start_time = time.time()

    def end_timer(self):
        self._end_time = time.time()

    def get_elapsed_time(self) -> float:
        return self._end_time - self._start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.retry = retry
        self._attempt: int = 1

    def get_max_attempt(self) -> int:
        return self.retry + 1

    def get_attempt(self) -> int:
        return self._attempt

    def increase_attempt(self):
        self._attempt += 1

    def reset_attempt(self):
        self._attempt = 0

    def should_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt <= max_attempt

    def is_last_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self._is_done: bool = False

    def mark_as_done(self):
        self._is_done = True

    def mark_as_undone(self):
        self._is_done = False

    def is_done(self) -> bool:
        return self._is_done


@typechecked
class PidModel():

    def __init__(self):
        self.zrb_task_pid: int = os.getpid()

    def set_task_pid(self, pid: int):
        self.zrb_task_pid = pid

    def get_task_pid(self) -> int:
        return self.zrb_task_pid


@typechecked
class TaskDataModel():

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
    ):
        self.name = name
        self.group = group
        self.envs = envs
        self.icon = icon
        self.color = color
        self._input_map: Mapping[str, Any] = {}
        self._env_map: Mapping[str, str] = {}
        self._is_keyval_set = False  # Flag

    def get_icon(self) -> str:
        if self.icon is None or self.icon == '':
            self.icon = get_random_icon()
        return self.icon

    def get_color(self) -> str:
        if self.color is None or not is_valid_color(self.color):
            self.color = get_random_color()
        return self.color

    def get_cmd_name(self) -> str:
        return get_cmd_name(self.name)

    def log_debug(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix}: {message}', attrs=['dark']
        )
        logging.debug(colored_message)

    def log_warn(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix}: {message}', attrs=['dark']
        )
        logging.warning(colored_message)

    def log_info(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix}: {message}', attrs=['dark']
        )
        logging.info(colored_message)

    def log_error(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix}: {message}', color='red', attrs=['bold']
        )
        logging.error(colored_message, exc_info=True)

    def print_out(self, msg: Any):
        prefix = self._get_colored_log_prefix()
        print(f'🤖 ➜ {prefix} • {msg}'.rstrip())

    def print_err(self, msg: Any):
        prefix = self._get_colored_log_prefix()
        print(f'🤖 ⚠ {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def play_bell(self):
        print('\a')

    def _get_colored_log_prefix(self) -> str:
        return self.colored(self._get_log_prefix())

    def _get_log_prefix(self) -> str:
        '''
        Return log prefix representing current task.
        This implementation override AccessoriesModel.get_log_prefix.
        '''
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        now = datetime.datetime.now().isoformat()
        pid = self.get_task_pid()
        info = f'{now} ⚙ {pid} ➤ {attempt} of {max_attempt}'
        icon = self.get_icon()
        name = self._get_complete_name()
        filled_name = name.rjust(13, ' ')
        return f'{info} • {icon} {filled_name}'

    def get_input_map(self) -> Mapping[str, Any]:
        return self._input_map

    def get_env_map(self) -> Mapping[str, str]:
        return self._env_map

    def inject_env_map(
        self, env_map: Mapping[str, str], override: bool = False
    ):
        for key, val in env_map.items():
            if override or key not in self._env_map:
                self._env_map[key] = val

    def get_float(self, val: Union[str, float]) -> float:
        if isinstance(val, str):
            return float(self.render_str(val))
        return val

    def get_int(self, val: Union[str, int]) -> int:
        if isinstance(val, str):
            return int(self.render_str(val))
        return val

    def render_str(self, val: str) -> str:
        template = jinja2.Template(val)
        data = self._get_default_render_data()
        self.log_debug(f'Render string template: {val}\nWith data: {data}')
        rendered_text = template.render(data)
        self.log_debug(f'Rendered result: {rendered_text}')
        return rendered_text

    def render_file(self, location: str) -> str:
        location_dir = os.path.dirname(location)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(location)
        data = self._get_default_render_data()
        data['TEMPLATE_DIR'] = location_dir
        self.log_debug(f'Render template: {location}\nWith data: {data}')
        rendered_text = template.render(data)
        self.log_debug(f'Rendered result: {rendered_text}')
        return rendered_text

    def _get_default_render_data(self) -> Mapping[str, Any]:
        return {
            'env': get_object_from_keyval(self._env_map),
            'input': get_object_from_keyval(self._input_map),
        }

    def set_local_keyval(
        self,
        input_map: Mapping[str, Any],
        env_prefix: str = ''
    ):
        if self._is_keyval_set:
            return True
        self._is_keyval_set = True
        self._input_map = dict(input_map)
        self.log_debug(f'Input map: {self._input_map}')
        self._env_map = os.environ
        for task_env in self.envs:
            env_name = task_env.name
            env_value = task_env.get(env_prefix)
            self._env_map[env_name] = env_value
        self.log_debug(f'Env map: {self._env_map}')

    def _get_complete_name(self):
        cmd_name = self.get_cmd_name()
        if self.group is None:
            return cmd_name
        group_cmd_name = self.group.get_complete_name()
        return f'{group_cmd_name} {cmd_name}'

    def show_celebration(self):
        complete_name = self._get_complete_name()
        elapsed_time = self.get_elapsed_time()
        icon = self.get_icon()
        print('🤖 🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉')
        print(self.colored(f'🤖 {icon} {complete_name} completed in'))
        print(self.colored(f'🤖 {icon} {elapsed_time} seconds'))
        print('🤖 🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉')
        self.play_bell()


@typechecked
class TaskModel(
    TaskDataModel, PidModel, FinishTracker, AttemptTracker, TimeTracker
):

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
    ):
        TaskDataModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            icon=icon,
            color=color
        )
        PidModel.__init__(self)
        FinishTracker.__init__(self)
        AttemptTracker.__init__(self, retry=retry)
        TimeTracker.__init__(self)



from zrb.task.code_maker import CodeMaker
import os
import pathlib
import shutil


def test_code_maker():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    template_path = os.path.join(dir_path, 'template')
    destination_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(destination_path):
        shutil.rmtree(destination_path)
    # initiate code_maker
    code_maker = CodeMaker(
        name='create-app',
        template_path=template_path,
        destination_path=destination_path,
        excludes=['*/excluded'],
        replacements={
            'app_name': 'my_app',
            '3000': '8080'
        }
    )
    main_loop = code_maker.create_main_loop()
    result = main_loop()
    assert result
    # excluded should not exists
    assert not os.path.exists(os.path.join(
        destination_path, 'excluded'
    ))
    # config_my_app.py should exists and contain the right data
    assert read_file(
        os.path.join(destination_path, 'src', 'config_my_app.py')
    ) == '\n'.join([
        'port = 8080',
        ''
    ])
    # my_app.py should exists and contain the right data
    assert read_file(
        os.path.join(destination_path, 'src', 'my_app.py')
    ) == '\n'.join([
        'from config_my_app import port',
        '',
        '',
        'def start():',
        "    print(f'starting my_app on port {port}')",
        '    return True',
        ''
    ])


def read_file(file_name: str) -> str:
    with open(file_name) as f:
        return f.read()



from ._group import show_group
from ...task.cmd_task import CmdTask

show_solid_principle = CmdTask(
    name='solid-principle',
    group=show_group,
    cmd=[
        'echo S - Single Responsibility Principle',
        'echo O - Open/Closed Principle',
        'echo L - Liskov’s Substitution Principle',
        'echo I - Interface Segregation Principle',
        'echo D - Dependency Inversion Principle'
    ]
)



from .runner import runner
from .task.base_task import BaseTask
from .task.task import Task
from .task.cmd_task import CmdTask
from .task.http_checker import HTTPChecker
from .task.port_checker import PortChecker
from .task.code_maker import CodeMaker
from .task_input.str_input import StrInput
from .task_env.env import Env
from .task_group.group import Group
